@@ -69,7 +69,7 @@ run_enrichment <- function(object, view, feature.sets, factors = "all",
   # if (statistical.test %in% c("cor.adj.parametric")) {
   idx <- apply(data,2, function(x) var(x,na.rm=TRUE))==0
   if (sum(idx)>=1) {
-    warning(sprintf("%d fetures were removed because they had no variance in the data.\n",sum(idx)))
+    warning(sprintf("%d features were removed because they had no variance in the data.\n",sum(idx)))
     data <- data[,!idx]
     W <- W[!idx,]
   }
@@ -180,6 +180,14 @@ run_enrichment <- function(object, view, feature.sets, factors = "all",
   if(!p.adj.method %in%  p.adjust.methods) 
     stop("p.adj.method needs to be an element of p.adjust.methods")
   adj.p.values <- apply(results$p.values, 2,function(lfw) p.adjust(lfw, method = p.adj.method))
+
+  # If we specify a direction, we are only interested in overrepresented pathawys in the selected direction
+  if (sign%in%c("positive","negative")) {
+    results$p.values[results$statistics<0] <- 1.0
+    adj.p.values[results$statistics<0] <- 1.0
+    results$statistics[results$statistics<0] <- 0
+  }
+  
   
   # If we specify a direction, we are only interested in overrepresented pathawys in the selected direction
   if (sign%in%c("positive","negative")) {
@@ -536,6 +544,7 @@ plot_enrichment_detailed <- function(enrichment.results, factor,
       mean.diff <- mean(pc.feature.stats[indexes.for.feature.set],na.rm=TRUE) - mean(pc.feature.stats[not.set.indexes], na.rm=TRUE)
       # compute the pooled standard deviation
       pooled.sd <- sqrt(((m1-1)*var(pc.feature.stats[indexes.for.feature.set], na.rm=TRUE) + (m2-1)*var(pc.feature.stats[not.set.indexes], na.rm=TRUE))/(m1+m2-2))
+
       # compute the t-statistic
       if (cor.adjustment) {
         t.stat <- mean.diff/(pooled.sd*sqrt(vif/m1 + 1/m2))
