@@ -1,4 +1,3 @@
-
 ###########################################
 ## Functions to visualise the input data ##
 ###########################################
@@ -246,7 +245,7 @@ plot_data_scatter <- function(object, factor = 1, view = 1, groups = "all", feat
     Y <- do.call(cbind, object@data[[view]][groups])
   }
   
-  # Feetch factors
+  # Fetch factors
   Z <- get_factors(object, factors = factor, groups = groups, as.data.frame = TRUE)
   Z <- Z[,c("sample","value")]
   colnames(Z) <- c("sample","x")
@@ -273,8 +272,7 @@ plot_data_scatter <- function(object, factor = 1, view = 1, groups = "all", feat
     stop("Features need to be either a numeric or character vector")
   }
   W <- W[features]
-  Y <- Y[features,,drop = FALSE]
-  
+
   # Set group/color/shape
   if (length(color_by)==1 & is.character(color_by)) color_name <- color_by
   if (length(shape_by)==1 & is.character(shape_by)) shape_name <- shape_by
@@ -287,7 +285,11 @@ plot_data_scatter <- function(object, factor = 1, view = 1, groups = "all", feat
   
   # Create data frame 
   foo <- list(features); names(foo) <- view
-  df2 <- get_data(object, groups = groups, features = foo, as.data.frame = TRUE)
+  if (isTRUE(imputed)) {
+    df2 <- get_imputed_data(object, groups = groups, views = view, features = foo, as.data.frame = TRUE)
+  } else {
+    df2 <- get_data(object, groups = groups, features = foo, as.data.frame = TRUE)
+  }
   df2$sample <- as.character(df2$sample)
   df <- dplyr::left_join(df1, df2, by = "sample")
   
@@ -339,19 +341,22 @@ plot_data_scatter <- function(object, factor = 1, view = 1, groups = "all", feat
 }
 
 
-
 #' @title Overview of the input data
 #' @name plot_data_overview
 #' @description Function to do a tile plot showing the missing value structure of the input data
 #' @param object a \code{\link{MOFA}} object.
+#' @param covariate (only for MEFISTO) specifies sample covariate to order samples by in the plot. This should be
+#' a character or  a numeric index giving the name or position of a column present in the covariates slot of the object.
+#' Default is the first sample covariate in covariates slot. \code{NULL} does not order by covariate
 #' @param colors a vector specifying the colors per view (see example for details).
+#' @param show_covariate (only for MEFISTO) boolean specifying whether to include the covariate in the plot
 #' @param show_dimensions logical indicating whether to plot the dimensions of the data (default is TRUE).
 #' @details This function is helpful to get an overview of the structure of the data. 
 #' It shows the model dimensionalities (number of samples, groups, views and features) 
 #' and it indicates which measurements are missing.
 #' @import ggplot2
 #' @importFrom reshape2 melt
-#' @importFrom dplyr mutate
+#' @importFrom dplyr mutate left_join
 #' @return A \code{\link{ggplot}} object
 #' @export
 #' @examples
@@ -360,7 +365,7 @@ plot_data_scatter <- function(object, factor = 1, view = 1, groups = "all", feat
 #' model <- load_model(file)
 #' plot_data_overview(model)
 
-plot_data_overview <- function(object, colors = NULL, show_dimensions = TRUE) {
+plot_data_overview <- function(object, covariate = 1, colors = NULL, show_covariate = FALSE, show_dimensions = TRUE) {
   
   # Sanity checks
   if (!is(object, "MOFA")) stop("'object' has to be an instance of MOFA")
@@ -368,6 +373,19 @@ plot_data_overview <- function(object, colors = NULL, show_dimensions = TRUE) {
   # Collect relevant data
   data <- object@data
   
+  # Collect covariate
+  if(any(object@dimensions[["C"]] < 1, is.null(object@covariates))) 
+    covariate <- NULL
+  if(!is.null(covariate)){
+    if(is.numeric(covariate)){
+      if(covariate > object@dimensions[["C"]]) stop("Covariate index out of range")
+      covariate <- covariates_names(object)[covariate]
+    }
+    if(!is.character(covariate) | !covariate %in% covariates_names(object)) 
+      stop("Covariate mispecified. Please read the documentation")
+    covari <- .set_xax(object, covariate)
+  }
+  
   M <- get_dimensions(object)[["M"]]
   G <- get_dimensions(object)[["G"]]
   if (M==1 & G==1) warning("This function is not useful when there is just one view and one group")
@@ -403,7 +421,12 @@ plot_data_overview <- function(object, colors = NULL, show_dimensions = TRUE) {
 
   # Melt to data.frame
   to.plot <- melt(ovw, id.vars = c("sample", "group"), var=c("view"))
-  to.plot$sample <- factor(to.plot$sample, levels = rownames(ovw))
+  if(!is.null(covariate)) {
+    to.plot <- left_join(to.plot, covari, by= "sample")
+    to.plot$sample <- factor(to.plot$sample, levels = unique(to.plot$sample[order(to.plot$covariate_value)]))
+  } else {
+    to.plot$sample <- factor(to.plot$sample, levels = rownames(ovw))
+  }
 
   n <- length(unique(to.plot$sample))
   
@@ -420,7 +443,10 @@ plot_data_overview <- function(object, colors = NULL, show_dimensions = TRUE) {
   } else {
     to.plot <- mutate(to.plot, view_label = view, group_label = group)
   }
-    
+  
+  # Order groups  
+  to.plot$group_label <- factor(to.plot$group_label, levels=unique(to.plot$group_label))
+  
   # Plot
   p <- ggplot(to.plot, aes_string(x="sample", y="view_label", fill="combi")) +
     geom_tile() +
@@ -440,6 +466,24 @@ plot_data_overview <- function(object, colors = NULL, show_dimensions = TRUE) {
       panel.grid = element_blank()
     )
   
+  if(show_covariate){
+    p2 <- ggplot(to.plot, aes_string(x="sample", y="covariate_value")) +
+      geom_point(size = 0.5) +  theme_bw() +theme(
+        text = element_text(size=10),
+        axis.ticks.x = element_blank(),
+        axis.title.x = element_blank(),
+        axis.text.x = element_blank(),
+        strip.background = element_blank(),
+        strip.text = element_blank()
+      ) + ylab(covariate) + facet_wrap(~group_label, ncol =1, scales="free_x")
+    
+    if(object@dimensions["G"] == 1) {
+    p <- cowplot::plot_grid(p, p2, align = "v", ncol = 1, rel_heights = c(1,0.2) )
+    } else{
+      p <- cowplot::plot_grid(p, p2, align = "h", nrow = 1, rel_widths = c(1,1) )
+    }
+  }
+  
   return(p)
 }
 
