@@ -72,6 +72,7 @@
       Z <- get_factors(object, groups=g, factors=k)[[1]][,1]
       Z <- Z[!is.na(Z)]
       
+      warning("Outlier detection is independent of the inferred lengthscale currently - might lead to unwanted results")
       cutoff <- 2.5 * 1.96
       tmp <- abs(Z - mean(Z)) / sd(Z)
 
@@ -118,6 +119,21 @@ return(model)
   
 }
 
+.check_and_get_covariates <- function(object, covariates) {
+  stopifnot(!any(duplicated(covariates)))
+  if (is.numeric(covariates)) {
+    stopifnot(all(covariates <= object@dimensions$C))
+    covariates_names(object)[covariates] 
+  } else {
+    if (paste0(covariates, collapse = "") == "all") { 
+      covariates_names(object)
+    } else {
+      stopifnot(all(covariates %in% covariates_names(object)))
+      covariates
+    }
+  }
+}
+
 .check_and_get_views <- function(object, views, non_gaussian=TRUE) {
   stopifnot(!any(duplicated(views)))
   if (is.numeric(views)) {
@@ -205,7 +221,8 @@ return(model)
   nodes_types <- list(
     multiview_nodes  = c("W", "AlphaW", "ThetaW"),
     multigroup_nodes = c("Z", "AlphaZ", "ThetaZ"),
-    twodim_nodes     = c("Y", "Tau")
+    twodim_nodes     = c("Y", "Tau"),
+    multivariate_singleview_node = "Sigma"
   )
 }
 
@@ -295,6 +312,44 @@ setReplaceMethod("colnames", signature(x = "matrix_placeholder"),
   return(df)
 }
 
+# (Hidden) function to define the color
+.set_xax <- function(object, xax) {
+  
+    # Option 1: by a metadata column in object@samples_metadata
+  if ((length(xax) == 1) && (is.character(xax)|is.factor(xax)) & (xax[1] %in% colnames(samples_metadata(object)))) {
+    xax <- samples_metadata(object)[,xax]
+    
+    # Option 2: by a feature present in the training data    
+  } else if ((length(xax) == 1) && is.character(xax) && (xax[1] %in% unlist(features_names(object)))) {
+    data <- lapply(get_data(object), function(l) Reduce(cbind, l))
+    features <- lapply(data, rownames)
+    viewidx <- which(sapply(features, function(x) xax %in% x))
+    xax <- data[[viewidx]][xax,]
+    
+    # Option 5: input is a data.frame with columns (sample, value)
+  } else if (is(xax, "data.frame")) {
+    stopifnot(all(colnames(xax) %in% c("sample", "value")))
+    stopifnot(all(unique(xax$sample) %in% unlist(samples_names(object))))
+    xax <- dplyr::rename(xax, covariate_value = value)
+    # Option 6: color_by is a vector of length N
+  } else if (length(xax) > 1) {
+    stopifnot(length(xax) == sum(get_dimensions(object)$N))
+    
+    # Option not recognised
+  } else {
+    stop("'xax' was specified but it was not recognised, please read the documentation")
+  }
+  
+  # Create data.frame with columns (sample,color)
+  if (!is(xax,"data.frame")) {
+    xax = data.frame(
+      sample = unlist(samples_names(object)),
+      covariate_value = xax,
+      stringsAsFactors = FALSE
+    )
+  }
+  return(xax)
+}
 
 # (Hidden) function to define the color
 .set_colorby <- function(object, color_by) {
@@ -482,3 +537,81 @@ setReplaceMethod("colnames", signature(x = "matrix_placeholder"),
     stroke <- 0.05
   }
 }
+
+# # (Hidden) function to define the shape
+# .set_shapeby_features <- function(object, shape_by, view) {
+#   
+#   # Option 1: no color
+#   if (is.null(shape_by)) {
+#     shape_by <- rep("1",sum(object@dimensions[["D"]][view]))
+#     
+#     # Option 2: input is a data.frame with columns (feature,color)
+#   } else if (is(shape_by,"data.frame")) {
+#     stopifnot(all(colnames(shape_by) %in% c("feature","color")))
+#     stopifnot(all(unique(shape_by$feature) %in% features(object)[[view]]))
+#     
+#     # Option 3: by a feature_metadata column
+#   } else if ((length(shape_by)==1) && is.character(shape_by) & (shape_by %in% colnames(features_metadata(object)))) {
+#     tmp <- features_metadata(object)
+#     shape_by <- tmp[tmp$view==view,shape_by]
+#     
+#     # Option 4: shape_by is a vector of length D
+#   } else if (length(shape_by) > 1) {
+#     stopifnot(length(shape_by) == object@dimensions[["D"]][[view]])
+#     
+#     # Option not recognised
+#   } else {
+#     stop("'shape_by' was specified but it was not recognised, please read the documentation")
+#   }
+#   
+#   # Create data.frame with columns (feature,shape)
+#   if (!is(shape_by,"data.frame")) {
+#     df = data.frame(
+#       feature = features(object)[[view]],
+#       shape_by = shape_by,
+#       view = view
+#     )
+#   }
+#   
+#   return(df)
+# }
+# 
+# 
+# # (Hidden) function to define the color
+# .set_colorby_features <- function(object, color_by, view) {
+#   
+#   # Option 1: no color
+#   if (is.null(color_by)) {
+#     color_by <- rep("1",sum(object@dimensions[["D"]][view]))
+#     
+#     # Option 2: input is a data.frame with columns (feature,color)
+#   } else if (is(color_by,"data.frame")) {
+#     stopifnot(all(colnames(color_by) %in% c("feature","color")))
+#     stopifnot(all(unique(color_by$feature) %in% features(object)[[view]]))
+#     
+#     # Option 3: by a feature_metadata column
+#   } else if ((length(color_by)==1) && is.character(color_by) & (color_by %in% colnames(features_metadata(object)))) {
+#     tmp <- features_metadata(object)
+#     color_by <- tmp[tmp$view==view,color_by]
+#     
+#     # Option 4: color_by is a vector of length D
+#   } else if (length(color_by) > 1) {
+#     stopifnot(length(color_by) == object@dimensions[["D"]][[view]])
+#     
+#     # Option not recognised
+#   } else {
+#     stop("'color_by' was specified but it was not recognised, please read the documentation")
+#   }
+#   
+#   # Create data.frame with columns (feature,color)
+#   if (!is(color_by,"data.frame")) {
+#     df = data.frame(
+#       feature = features(object)[[view]],
+#       color_by = color_by,
+#       view = view
+#     )
+#   }
+#   
+#   return(df)
+# }
+
